# -*- coding: utf-8 -*-

# standard library imports
import os
import statistics
import sys

# third-party imports
import click
import dask.bag as db
import gffpandas.gffpandas as gffpd
import numpy as np
import pandas as pd
from Bio import SeqIO
from dask.diagnostics import ProgressBar

# first-party imports
from loguru import logger

# module imports
from . import cli
from . import click_loguru
from .common import *
from .core import cluster_set_name
from .core import usearch_cluster

# global constants
GFF_EXT = "gff3"
FASTA_EXT = "faa"
HOMOLOGY_ENDING = "-homology.tsv"
FILES_ENDING = "-files.tsv"
SYNTENY_ENDING = "-synteny.tsv"
PROXY_ENDING = "-proxy.tsv"


def kmer_synteny_block_func(k, frame):
    """Return a synteny block closure and its name."""
    frame_len = len(frame)
    frame_iter = iter(range(frame_len))
    null_return = (
        0,
        0,
        0,
    )

    def kmer_synteny():
        """Mappable function from index to (footprint, direction, hash) tuple."""
        first_index = next(frame_iter)
        cluster_list = []
        initial_seq_id = frame.loc[first_index, "seq_id"]
        for idx in range(first_index, first_index + k):
            if idx + 1 > frame_len:
                return null_return
            cluster_size = frame.loc[idx, "cluster_size"]
            if frame.loc[idx, "seq_id"] != initial_seq_id or cluster_size == 1:
                return null_return
            cluster_list.append(cluster)
        fwd_hash = hash(tuple(cluster_list))
        rev_hash = hash(tuple(reversed(cluster_list)))
        if fwd_hash > rev_hash:
            hashval = fwd_hash
            direction = 1
        else:
            hashval = rev_hash
            direction = -1
        return k, direction, hashval

    return kmer_synteny, f"kmer{k}synteny"


def read_files(setname, synteny=None):
    """Read previously-calculated homology/synteny files and file frame."""
    set_path = Path(setname)
    files_frame_path = set_path / f"{setname}{FILES_ENDING}"
    try:
        file_frame = pd.read_csv(files_frame_path, index_col=0, sep="\t")
    except:
        logger.error(f"Unable to read files frome from {files_frame_path}")
        sys.exit(1)
    if synteny is None:
        ending = HOMOLOGY_ENDING
        file_type = "homology"
    else:
        ending = f"-{synteny}{SYNTENY_ENDING}"
        file_type = "synteny"
    paths = [p for p in set_path.glob("*" + ending)]
    stems = [p.name[: -len(ending)] for p in paths]
    if len(stems) != len(file_frame):
        logger.error(
            f"Number of {file_type} files ({len(stems)})is not the same as length of file frame({len(file_frame)})."
        )
        sys.exit(1)
    frame_dict = {}
    for i, path in enumerate(paths):
        logger.debug(f"Reading homology file {path}.")
        frame_dict[stems[i]] = pd.read_csv(path, index_col=0, sep="\t")
    return file_frame, frame_dict


@cli.command()
@click_loguru.init_logger()
@click.option("--identity", "-i", default=0.0, help="Minimum sequence ID (0-1). [default: lowest]")
@click.option("--clust/--no-clust", "-c/-x", is_flag=True, default=True, help="Do cluster calc.", show_default=True)
@click.option(
    "-s",
    "--shorten_source",
    default=False,
    is_flag=True,
    show_default=True,
    help="Remove invariant dotpaths in source IDs.",
)
@click.argument("setname")
@click.argument("gff_faa_path_list", nargs=-1)
def annotate_homology(identity, clust, shorten_source, setname, gff_faa_path_list):
    """Marshal homology and sequence information.

    Corresponding GFF and FASTA files must have a corresponding prefix to their
    file names, but theu may occur in any order in the list.  Paths to files
    need not be the same.  Files must be uncompressed. FASTA files must be
    protein files with extension ".faa".  GFF files must have extension ".gff3".

    IDs must correspond between GFF and FASTA files and must be unique across
    the entire set.
    """
    if not len(gff_faa_path_list):
        logger.error("No files in list, exiting.")
        sys.exit(0)
    # find matching pairs of gff files
    file_dict = {}
    gff_stems = [str(Path(n).stem) for n in gff_faa_path_list if n.find(GFF_EXT) > -1]
    gff_stems.sort(key=lambda x: len(x))
    fasta_stems = [str(Path(n).stem) for n in gff_faa_path_list if n.find(FASTA_EXT) > -1]
    fasta_stems.sort(key=lambda x: len(x))
    if len(gff_stems) != len(fasta_stems):
        logger.error(f"Differing number of {FASTA_EXT} ({len(fasta_stems)}) and {GFF_EXT} files ({len(gff_stems)}).")
        sys.exit(1)
    for fasta in fasta_stems:
        prefix_len = max([len(os.path.commonprefix([fasta, gff])) for gff in gff_stems])
        match_gff_idx = [i for i, gff in enumerate(gff_stems) if len(os.path.commonprefix([fasta, gff])) == prefix_len][
            0
        ]
        match_gff = gff_stems.remove(match_gff_idx)
        fasta_path = [Path(p) for p in gff_faa_path_list if p.endswith(fasta + "." + FASTA_EXT)][0]
        gff_path = [Path(p) for p in gff_faa_path_list if p.endswith(match_gff + "." + GFF_EXT)][0]
        stem = os.path.commonprefix([fasta, match_gff])
        file_dict[stem] = {GFF_EXT: gff_path, FASTA_EXT: fasta_path}
    frame_dict = {}
    set_path = Path(setname)
    set_path.mkdir(parents=True, exist_ok=True)
    # TODO-all combinations of file sets
    fasta_records = []
    for stem in file_dict.keys():
        logger.debug(f"Reading GFF file {file_dict[stem][GFF_EXT]}.")
        annotation = gffpd.read_gff3(file_dict[stem][GFF_EXT])
        mRNAs = annotation.filter_feature_of_type(["mRNA"]).attributes_to_columns()
        mRNAs.drop(
            mRNAs.columns.drop(["seq_id", "start", "strand", "ID"]), axis=1, inplace=True
        )  # drop non-essential columns
        if shorten_source:
            # drop identical sub-fields in seq_id to keep them visually short (for development)
            split_sources = mRNAs["seq_id"].str.split(".", expand=True)
            split_sources = split_sources.drop(
                [i for i in split_sources.columns if len(set(split_sources[i])) == 1], axis=1
            )
            sources = split_sources.agg(".".join, axis=1)
            mRNAs["seq_id"] = sources
        # TODO-calculated subfragments from repeated
        # TODO-sort GFFs in order of longest fragments
        # TODO-add gene order
        file_dict[stem]["fragments"] = len(set(mRNAs["seq_id"]))
        logger.debug(f"Reading FASTA file {file_dict[stem][FASTA_EXT]}.")
        fasta_dict = SeqIO.to_dict(SeqIO.parse(file_dict[stem][FASTA_EXT], "fasta"))
        # TODO-filter out crap and calculate ambiguous
        file_dict[stem]["n_seqs"] = len(fasta_dict)
        file_dict[stem]["residues"] = sum([len(fasta_dict[k].seq) for k in fasta_dict.keys()])
        mRNAs = mRNAs[mRNAs["ID"].isin(fasta_dict.keys())]
        mRNAs["protein_len"] = mRNAs["ID"].map(lambda k: len(fasta_dict[k].seq))
        frame_dict[stem] = mRNAs.set_index("ID")
        del annotation
        for key in fasta_dict.keys():
            fasta_records.append(fasta_dict[key])
    file_frame = pd.DataFrame.from_dict(file_dict).transpose()
    file_frame = file_frame.sort_values(by=["n_seqs"])
    file_frame["n"] = range(len(file_frame))
    file_frame["stem"] = file_frame.index
    file_frame = file_frame.set_index("n")
    logger.debug("Writing files frame.")
    file_frame.to_csv(set_path / f"{setname}{FILES_ENDING}", sep="\t")
    del file_dict
    set_keys = list(file_frame["stem"])
    concatenated_fasta_name = f"{setname}.faa"
    if clust:
        logger.debug(f"Writing concatenated FASTA file {concatenated_fasta_name}.")
        with (set_path / concatenated_fasta_name).open("w") as concat_fh:
            SeqIO.write(fasta_records, concat_fh, "fasta")
        logger.debug("Doing cluster calculation.")
        cwd = Path.cwd()
        os.chdir(set_path)
        stats, graph, hist, any, all = usearch_cluster.callback(
            concatenated_fasta_name, identity, write_ids=True, delete=False
        )
        os.chdir(cwd)
        del stats, graph, hist, any, all
    del fasta_records
    cluster_frame = pd.read_csv(set_path / (cluster_set_name(setname, identity) + "-ids.tsv"), sep="\t")
    cluster_frame = cluster_frame.set_index("id")
    logger.debug("Mapping FASTA IDs to cluster properties.")

    def id_to_cluster_property(id, column):
        try:
            return int(cluster_frame.loc[id, column])
        except KeyError:
            raise KeyError(f"ID {id} not found in clusters")

    for stem in set_keys:
        frame = frame_dict[stem]
        frame["cluster_id"] = frame.index.map(lambda id: id_to_cluster_property(id, "cluster"))
        frame["cluster_size"] = frame.index.map(lambda id: id_to_cluster_property(id, "siz"))
        homology_filename = f"{stem}{HOMOLOGY_ENDING}"
        logger.debug(f"Writing homology file {homology_filename}")
        frame.to_csv(set_path / homology_filename, sep="\t")


@cli.command()
@click_loguru.init_logger()
@click.option("-k", default=6, help="Synteny block length.", show_default=True)
@click.option("-r", "--rmer", default=False, is_flag=True, show_default=True, help="Allow repeats in block.")
@click.argument("setname")
def kmer_synteny(k, rmer, setname):
    """Calculate k-mer syntenic blocks among sets of GFF/FASTA files.
    """
    set_path = Path(setname)
    files_frame, frame_dict = read_files(setname)
    set_keys = list(files_frame["stem"])
    logger.debug(f"Calculating k-mer of length {k} synteny blocks.")
    merge_frame_columns = ["hash", "source"]
    merge_frame = pd.DataFrame(columns=merge_frame_columns)
    for stem in set_keys:
        frame = frame_dict[stem]
        if rmer:
            map_func = None
            synteny_func_name = "None"
        else:
            map_func, synteny_func_name = kmer_synteny_block_func(k, frame)
        frame["footprint"] = 0
        frame["hashdir"] = 0
        frame[synteny_func_name] = 0
        frame_len = frame.shape[0]
        for idx in range(frame_len):
            footprint, hashdir, block = map_func()
            frame.iloc[idx, FOOTPRINT_POS] = footprint
            frame.iloc[idx, HASHDIR_POS] = hashdir
            frame.iloc[idx, SYNTENY_POS] = block
        # TODO:E values
        hash_series = frame.loc[:, synteny_func_name]
        assigned_hashes = hash_series[hash_series != 0]
        del hash_series
        n_assigned = len(assigned_hashes)
        logger.info(f"{stem} has {frame_len} proteins, {n_assigned} of which have {synteny_func_name} hashes,")
        hash_counts = assigned_hashes.value_counts()
        assigned_hash_frame = pd.DataFrame(columns=merge_frame_columns)
        assigned_hash_frame["hash"] = assigned_hashes.unique()
        assigned_hash_frame["source"] = stem
        n_non_unique = n_assigned - len(assigned_hash_frame)
        percent_non_unique = n_non_unique / n_assigned * 100.0
        logger.info(f"  of which {n_non_unique} ({percent_non_unique})% are non-unique.")
        merge_frame.append(assigned_hash_frame)
        del assigned_hash_frame
        # create self_count column in frame
        frame["self_count"] = 0
        for idx, row in frame[frame[synteny_func_name] != 0].iterrows():
            frame.loc[idx, "self_count"] = hash_counts.loc[row[synteny_func_name]]
        del hash_counts
    logger.debug(f"Calculating overlap of {len(merge_frame)} hash terms.")
    hash_counts = merge_frame["hash"].value_counts()
    merged_hash_frame = pd.DataFrame(index=merge_frame["hash"].unique(), columns=["count"])
    for idx, row in merged_hash_frame.iterrows():
        merged_hash_frame.loc[idx, "count"] = hash_counts.loc[row[synteny_func_name]]
    print(f"Merged_hash_frame={merged_hash_frame}")
    merged_hash_frame = merged_hash_frame[merged_hash_frame["count"] > 1]
    print(f"after dropping non-matching hashes, len = {len(merged_hash_frame)}")
    print(f"merged hash counts={hash_counts}")
    for stem in set_keys:
        synteny_name = f"{stem}-{synteny_func_name}{SYNTENY_ENDING}"
        logger.debug(f"Writing {synteny_func_name} synteny frame {synteny_name}.")
        frame_dict[stem].to_csv(set_path / synteny_name, sep="\t")


def dagchainer_id_to_int(id):
    """Accepts DAGchainer ids such as "cl1" and returns an integer."""
    if not id.startswith("cl"):
        raise ValueError(f"Invalid ID {id}.")
    id_val = id[2:]
    if not id_val.isnumeric():
        raise ValueError(f"Non-numeric ID value in {id}.")
    return int(id_val)


@cli.command()
@click_loguru.init_logger()
@click.argument("setname")
def dagchainer_synteny(setname):
    """Read DAGchainer synteny blocks into homology frames.

    IDs must correspond between DAGchainer files and homology blocks.
    Currently does not calculate DAGchainer synteny.
    """
    synteny_func_name = "dagchainer"
    set_path = Path(setname)
    logger.debug(f"Reading {synteny_func_name} synteny file.")
    synteny_frame = pd.read_csv(set_path / synteny_func_name / "clusters.tsv", sep="\t")
    synteny_frame["synteny_id"] = synteny_frame["cluster"].map(lambda id: dagchainer_id_to_int(id))
    synteny_frame = synteny_frame.drop(["cluster"], axis=1)
    cluster_counts = synteny_frame["synteny_id"].value_counts()
    synteny_frame["synteny_count"] = synteny_frame["synteny_id"].map(cluster_counts)
    synteny_frame = synteny_frame.sort_values(by=["synteny_count", "synteny_id"])
    synteny_frame = synteny_frame.set_index(["id"])
    files_frame, frame_dict = read_files(setname)
    set_keys = list(files_frame["stem"])

    def id_to_synteny_property(id, column):
        try:
            return int(synteny_frame.loc[id, column])
        except KeyError:
            return 0

    for stem in set_keys:
        homology_frame = frame_dict[stem]
        homology_frame["synteny_id"] = homology_frame.index.map(lambda x: id_to_synteny_property(x, "synteny_id"))
        homology_frame["synteny_count"] = homology_frame.index.map(lambda x: id_to_synteny_property(x, "synteny_count"))
        synteny_name = f"{stem}-{synteny_func_name}{SYNTENY_ENDING}"
        logger.debug(f"Writing {synteny_func_name} synteny frame {synteny_name}.")
        homology_frame.to_csv(set_path / synteny_name, sep="\t")


class ProxySelector(object):

    """Provide methods for downselection of proxy genes."""

    def __init__(self, frame, prefs):
        """Calculate any joint statistics from frame."""
        self.frame = frame
        self.prefs = prefs
        self.reasons = []
        self.drop_ids = []
        self.first_choice = prefs[0]
        self.first_choice_hits = 0
        self.first_choice_unavailable = 0
        self.cluster_count = 0

    def choose(self, chosen_one, cluster, reason, drop_non_chosen=True):
        """Make the choice, recording stats."""
        self.frame.loc[chosen_one, "reason"] = reason
        self.first_choice_unavailable += int(self.first_choice not in set(cluster["stem"]))
        self.first_choice_hits += int(cluster.loc[chosen_one, "stem"] == self.first_choice)
        non_chosen_ones = list(cluster.index)
        non_chosen_ones.remove(chosen_one)
        if drop_non_chosen:
            self.drop_ids += non_chosen_ones
        else:
            self.cluster_count += len(non_chosen_ones)

    def choose_by_preference(self, subcluster, cluster, reason, drop_non_chosen=True):
        """Choose in order of preference."""
        stems = subcluster["stem"]
        pref_idxs = [subcluster[stems == pref].index for pref in self.prefs]
        pref_lens = np.array([int(len(idx) > 0) for idx in pref_idxs])
        best_choice = np.argmax(pref_lens)  # first occurrance
        if pref_lens[best_choice] > 1:
            raise ValueError(f"subcluster {subcluster} is not unique w.r.t. genome {list(stems)[best_choice]}.")
        self.choose(pref_idxs[best_choice][0], cluster, reason, drop_non_chosen)

    def choose_by_length(self, subcluster, cluster, drop_non_chosen=True):
        """Return an index corresponding to the selected modal/median length."""
        counts = subcluster["protein_len"].value_counts()
        max_count = max(counts)
        if max_count > 1:  # repeated values exist
            max_vals = list(counts[counts == max(counts)].index)
            modal_cluster = subcluster[subcluster["protein_len"].isin(max_vals)]
            self.choose_by_preference(
                modal_cluster, cluster, f"mode{len(modal_cluster)}", drop_non_chosen=drop_non_chosen
            )
        else:
            lengths = list(subcluster["protein_len"])
            median_vals = [statistics.median_low(lengths), statistics.median_high(lengths)]
            median_pair = subcluster[subcluster["protein_len"].isin(median_vals)]
            self.choose_by_preference(median_pair, cluster, "median", drop_non_chosen=drop_non_chosen)

    def cluster_selector(self, cluster):
        "Calculate which gene in a homology cluster should be left and why."
        self.cluster_count += 1
        if len(cluster) == 1:
            self.choose(cluster.index[0], cluster, "singleton")
        else:
            for synteny_id, subcluster in cluster.groupby(by=["synteny_id"]):
                if len(subcluster) > 1:
                    self.choose_by_length(subcluster, cluster, drop_non_chosen=(not synteny_id))
                else:
                    if subcluster["synteny_id"][0] != 0:
                        self.choose(subcluster.index[0], cluster, "bad_synteny", drop_non_chosen=(not synteny_id))
                    else:
                        self.choose(subcluster.index[0], cluster, "single", drop_non_chosen=(not synteny_id))

    def downselect_frame(self):
        """Return a frame with reasons for keeping and non-chosen-ones dropped."""
        drop_pct = len(self.drop_ids) * 100.0 / len(self.frame)
        logger.info(f"Dropping {len(self.drop_ids)} ({drop_pct:0.1f}%) of {len(self.frame)} genes.")
        return self.frame.drop(self.drop_ids)

    def selection_stats(self):
        """Return selection stats."""
        return self.cluster_count, self.first_choice_unavailable, self.first_choice_hits


@cli.command()
@click_loguru.init_logger()
@click.argument("setname")
@click.argument("synteny_type")
@click.argument("prefs", nargs=-1)
def proxy_genes(setname, synteny_type, prefs):
    """Calculate a set of proxy genes from synteny files.

    prefs is an optional list of genome stems in order of preference in the proxy calc.
    """
    set_path = Path(setname)
    files_frame, frame_dict = read_files(setname, synteny=synteny_type)
    set_keys = list(files_frame["stem"])
    default_prefs = set_keys.copy()
    default_prefs.reverse()
    if prefs != ():
        for stem in prefs:
            if stem not in default_prefs:
                logger.error(f"Preference {stem} not in {default_prefs}")
                sys.exit(1)
            else:
                default_prefs.remove(stem)
        prefs = list(prefs) + default_prefs
        order = "non-default"
    else:
        prefs = default_prefs
        order = "default"
    logger.debug(f"Genome preference for proxy selection in {order} order: {prefs}")
    proxy_frame = None
    for stem in set_keys:
        logger.debug(f"Reading {stem}")
        frame_dict[stem]["stem"] = stem
        if proxy_frame is None:
            proxy_frame = frame_dict[stem]
        else:
            proxy_frame = proxy_frame.append(frame_dict[stem])
    del files_frame
    proxy_frame = proxy_frame.sort_values(by=["cluster_size", "cluster_id", "synteny_count", "synteny_id"])
    proxy_filename = f"{setname}-{synteny_type}{PROXY_ENDING}"
    logger.debug(f"Writing initial proxy file {proxy_filename}.")
    proxy_frame.to_csv(set_path / proxy_filename, sep="\t")
    proxy_frame["reason"] = ""
    logger.debug(f"Downselecting homology clusters.")
    downselector = ProxySelector(proxy_frame, prefs)
    for cluster_id, homology_cluster in proxy_frame.groupby(by=["cluster_id"]):
        downselector.cluster_selector(homology_cluster)
    downselected = downselector.downselect_frame()
    downselected_filename = f"{setname}-{synteny_type}-downselected{PROXY_ENDING}"
    logger.debug(f"Writing downselected proxy file {downselected_filename}.")
    downselected.to_csv(set_path / downselected_filename, sep="\t")
    # print out stats
    cluster_count, first_choice_unavailable, first_choice_hits = downselector.selection_stats()
    first_choice_percent = first_choice_hits * 100.0 / (cluster_count - first_choice_unavailable)
    first_choice_unavailable_percent = first_choice_unavailable * 100.0 / cluster_count
    logger.info(f"First-choice ({prefs[0]}) selections from {cluster_count} homology clusters:")
    logger.info(f"   not in cluster: {first_choice_unavailable} ({first_choice_unavailable_percent:.1f}%)")
    logger.info(f"   chosen as proxy: {first_choice_hits} ({first_choice_percent:.1f}%)")



# -*- coding: utf-8 -*-
"""
data analysis and plotting
"""
# standard library imports
import sys

# third-party imports
import click
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns

# first-party imports
from loguru import logger

# module imports
from . import cli
from . import click_loguru
from .common import ALLFILE_SUFFIX
from .common import ANYFILE_SUFFIX
from .common import STATFILE_SUFFIX

# Global constants
IDENT_LOG_MIN = -3
IDENT_LOG_MAX = 0
EPSILON = 0.000001
FILETYPE = "png"
MAX_BINS = 10


def make_histogram(dist, name, log10=False, bins=None):
    # do histogram plot with kernel density estimate
    dist = dist[dist < 10]
    mean = dist.mean()
    if log10:
        dist = np.log10(dist)
    # if len(dist) < MAX_BINS:
    #    bins = len(dist)
    # else:
    #    bins = MAX_BINS
    sns.distplot(
        dist,
        bins=None,
        rug=False,
        kde=False,
        norm_hist=True,
        rug_kws={"color": "b"},
        kde_kws={"color": "k", "linewidth": 1, "label": "KDE"},
        hist_kws={
            "histtype": "step",
            "linewidth": 2,
            "alpha": 1,
            "color": "b",
            #'range':(0,20)
        },
    )
    plt.title(f"{name} histogram of {len(dist):d} values, mean={mean:.1f}")
    if log10:
        plt.xlabel("log " + name)
    else:
        plt.xlabel(name)
    plt.ylabel("Frequency")
    # for ext in PLOT_TYPES:
    #    plt.savefig(LOG_PATH / ('%s-histogram.' % (name.rstrip('%')) + ext),
    #                bbox_inches='tight')
    plt.show()
    # plt.close('all')


def tick_function(X):
    X = X * 3.0 - 3
    vals = [(f"{v:f}").rstrip("0").rstrip(".") for v in (1.0 - 10 ** X) * 100.0]
    ticks = [f"{v}%" for v in vals]
    return ticks


def log_deriv(X, Y):
    logX = -1.0 * np.log10(X + EPSILON)
    logY = np.log10(Y)
    return np.gradient(logY) / np.gradient(logX)


@cli.command()
@click_loguru.init_logger()
@click.argument("instemlist")
def analyze_clusters(dirname, instemlist, label, reference=None, on_id=None, match_type=None):
    if match_type is None:
        matches = ["all", "any"]
    else:
        matches = [match_type]
    uniques = {}
    divergence = {}
    dirpath = Path(dirname)
    div_dist = {"all": {"ref": 0.0}, "any": {"ref": 0.0}}
    print("ref=", reference)
    for stem in instemlist:
        print("stem=", stem)
        paths = {
            "all": dirpath / (stem + ALLFILE_SUFFIX),
            "any": dirpath / (stem + ANYFILE_SUFFIX),
            "stat": dirpath / (stem + STATFILE_SUFFIX),
        }
        stats = pd.read_csv(paths["stat"], sep="\t", index_col=0)
        uniques[stem] = stats["unique_seqs"].iloc[0]
        divergence[stem] = stats["divergence"]
        if on_id is None:
            div_dist["all"][stem] = log_deriv(divergence[stem], stats["clusters"])
            div_dist["any"][stem] = None
            if stem == reference:
                div_dist["all"]["ref"] = div_dist["all"][stem]
                div_dist["any"]["ref"] = None
        else:
            for match in ["any", "all"]:
                data = pd.read_csv(paths[match], sep="\t", index_col=0)
                try:
                    div_dist[match][stem] = log_deriv(divergence[stem], data.loc[on_id])
                except KeyError:  # this label wasn't found
                    div_dist[match][stem] = None
                if stem == reference:
                    div_dist[match]["ref"] = div_dist[match][stem]
    #
    # Make the plots
    #
    plt.style.use("seaborn-whitegrid")
    axes = {}
    fig, ax = plt.subplots(len(matches), sharex=True)
    try:
        for axis, i in enumerate(ax):
            axes[matches[i]] = axis
            loweraxis = ax[1]
    except TypeError:
        axes[matches[0]] = ax
        loweraxis = ax
    for stem in instemlist:
        for match in matches:
            if div_dist[match][stem] is None:
                continue
            axes[match].plot(
                divergence[stem],
                div_dist[match][stem] - div_dist[match]["ref"],
                label="%s" % (stem.replace(label + ".", "")),
            )
            # uniques[stem]/1000.))
    if reference is None:
        if on_id is None:
            title = f"{label} Divergence Distribution"
            outfilestem = f"{label}_divergence_dist."
        else:
            title = f'{label} Divergence Distribution on "{on_id}"'
            outfilestem = f"{label}_divergence_dist_{on_id}."
    else:
        if on_id is None:
            title = f"{label}_Differential Divergence Distribution vs. {reference}"
            outfilestem = f"{label}_divergence_dist_vs{reference}."
        else:
            title = f'{label} Differential Divergence Distribution on "{on_id}" vs. {reference}'
            outfilestem = f"{label}_divergence_dist_on_{on_id}_vs_ref."
    if reference is None:
        fig.text(0.02, 0.5, "Logarithmic Derivative on Clusters", ha="center", va="center", rotation="vertical")
    else:
        fig.text(
            0.02, 0.5, "Logarithmic Derivative Difference on Clusters", ha="center", va="center", rotation="vertical"
        )
    if len(matches) == 2:
        fig.text(0.5, 0.47, "All in Cluster", ha="center", va="center")
        fig.text(0.5, 0.89, "Any in Cluster", ha="center", va="center")
    else:
        fig.text(0.5, 0.91, "%s in Cluster" % matches[0].capitalize(), ha="center", va="center")
    loweraxis.set(xlabel="Divergence on Sequence Identity")
    loweraxis.legend(loc="upper left")
    fig.suptitle(title)
    plt.xscale("log")
    limits = [0.001, 1.0]
    new_tick_locations = np.array([0.0, 1.0 / 3.0, 2.0 / 3.0, 1.0])
    loweraxis.set_xlim(limits)
    axes["second"] = loweraxis.twiny()
    axes["second"].set_xlim(limits)
    axes["second"].set_xticks(new_tick_locations)
    axes["second"].set_xticklabels(tick_function(new_tick_locations))
    axes["second"].set_xlabel("   ")
    # r'%Identity')
    # plt.ylim([-0.002,0.002])
    outfilename = outfilestem + f"{FILETYPE}"
    print(f"saving plot to {outfilename}")
    plt.savefig(dirpath / outfilename, dpi=200)
    plt.show()


def do_cuts(obs, high, low, label):
    if high > 0.0:
        hicuts = obs[obs > high]
        obs = obs[obs <= high]
        if len(hicuts):
            hifilename = label + "_hicuts.tsv"
            logger.info(
                "%d observations dropped by high-side cutoff of %.2f written to %s", len(hicuts), high, hifilename
            )
            logger.info(hicuts)
    if low > 0.0:
        locuts = obs[obs > low]
        obs = obs[obs >= low]
        logger.info("%d observations dropped by low-side cutoff of %.2f", len(locuts), low)
        if len(locuts):
            lofilename = label + "_locuts.tsv"
            logger.info(
                "%d observations dropped by low-side cutoff of %.2f written to %s", len(locuts), low, lofilename
            )
            logger.info(locuts)
    return obs


@cli.command()
@click_loguru.init_logger()
@click.option("--hi_cutoff", default=2.0, show_default=True, help="Disregard above this value.")
@click.option("--lo_cutoff", default=0.0, show_default=True, help="Disregard below this value.")
@click.argument("cluster_size")
@click.argument("combinedfile")
def outlier_length_dist(hi_cutoff, lo_cutoff, cluster_size, combinedfile):
    """Plot the normalized length distribution of singletons in clusters"""
    cluster_size = int(cluster_size)
    if not cluster_size > 0:
        logger.error("Positive cluster size must be specified")
        sys.exit(1)
    clusters = pd.read_csv(combinedfile, sep="\t", index_col=0)
    norm_lengths = []
    for cluster_id, cluster in clusters.groupby("cluster"):
        if cluster["siz"].iloc[0] != cluster_size:
            # not the right size
            continue
        if len(set(cluster["sub"])) != 2:
            # not just two subclusters
            continue
        if 1 not in set(cluster["sub_siz"]):
            # no singleton cluster
            continue
        singleton = cluster[cluster["sub_siz"] == 1]
        length = singleton["norm"].iloc[0]
        norm_lengths.append(length)
    norm_lengths = np.array(norm_lengths)
    norm_lengths = do_cuts(norm_lengths, hi_cutoff, lo_cutoff, "len")
    logger.info("%d singleton outliers in clusters of size %d", len(norm_lengths), cluster_size)
    logger.info("min:\t%.3f", min(norm_lengths))
    logger.info("max:\t%.3f", max(norm_lengths))
    logger.info("mean: %.3f", norm_lengths.mean())
    ax = sns.distplot(norm_lengths, bins=100, kde_kws={"label": "KDE"})
    ax.set_xlabel("Normalized Length of Singleton")
    plt.title("Length distribution of %d singleton subclusters" % (len(norm_lengths)))
    outfilename = f"norm_len_dist.{FILETYPE}"
    logger.info("saving plot to %s", outfilename)
    # plt.yscale('log')
    plt.savefig(outfilename, dpi=200)
    plt.show()


@cli.command()
@click_loguru.init_logger()
@click.option("--hi_cutoff", default=0.0, show_default=True, help="Disregard above this value.")
@click.option("--lo_cutoff", default=0.0, show_default=True, help="Disregard below this value.")
@click.argument("cluster_size")
@click.argument("combinedfile")
def length_std_dist(cluster_size, hi_cutoff, lo_cutoff, combinedfile):
    """Plot the normalized length distribution of singletons in clusters"""
    cluster_size = int(cluster_size)
    if not cluster_size > 0:
        logger.error("Positive cluster size must be specified")
        sys.exit(1)
    clusters = pd.read_csv(combinedfile, sep="\t", index_col=0)
    stds = []
    for cluster_id, cluster in clusters.groupby("cluster"):
        if cluster["siz"].iloc[0] != cluster_size:
            # not the right size
            continue
        if len(set(cluster["sub"])) != 1:
            # Only one subcluster
            continue
        val = cluster["std"].iloc[0]
        stds.append(val)
    stds = np.array(stds)
    pct_zeros = len(stds[stds == 0.0]) * 100 / len(stds)
    stds = do_cuts(stds, hi_cutoff, lo_cutoff, "stds")
    logger.info("%d single-subgroup clusters of size %d", len(stds), cluster_size)
    logger.info("%.1f %% zeroes, max is %.2f", pct_zeros, max(stds))
    logger.info("mean is %.3f", stds.mean())
    logbins = np.logspace(0.7, 3, 100)
    ax = sns.distplot(stds, bins=logbins, kde=False)
    ax.set_xlabel("Standard Deviation of Single-Subgroup Clusters")
    title = "Length Standard Deviation distribution of %d clusters" % len(stds)
    plt.title(title)
    outfilename = f"std_dist.{FILETYPE}"
    logger.info("saving plot to %s", outfilename)
    plt.yscale("log")
    plt.xscale("log")
    plt.savefig(outfilename, dpi=200)
    plt.show()



# -*- coding: utf-8 -*-
"""
Constants and functions in common across modules
"""
# standard library imports
from pathlib import Path

# first-party imports
import poetry_version

NAME = "azulejo"
VERSION = "".join(poetry_version.extract(source_file=__file__))
STATFILE_SUFFIX = f"-{NAME}_stats.tsv"
ANYFILE_SUFFIX = f"-{NAME}_ids-any.tsv"
ALLFILE_SUFFIX = f"-{NAME}_ids-all.tsv"
CLUSTFILE_SUFFIX = f"-{NAME}_clusts.tsv"
SEQ_FILE_TYPE = "fasta"


def get_paths_from_file(f, must_exist=True):
    inpath = Path(f).expanduser().resolve()
    if must_exist and not inpath.exists():
        raise FileNotFoundError(f)
    dirpath = inpath.parent
    return inpath, dirpath



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['azulejo']

package_data = \
{'': ['*'], 'azulejo': ['bin/*']}

install_requires = \
['biopython>=1.76,<2.0',
 'click>=7.1.2,<8.0.0',
 'click_loguru>=0.1.0,<0.2.0',
 'click_plugins>=1.1.1,<2.0.0',
 'coverage>=5.1,<6.0',
 'dask[bag]>=2.15.0,<3.0.0',
 'flynt>=0.46.1,<0.47.0',
 'gffpandas>=1.2.0,<2.0.0',
 'loguru>=0.4.1,<0.5.0',
 'matplotlib>=3.2.1,<4.0.0',
 'networkx>=2.4,<3.0',
 'numpy>=1.18.3,<2.0.0',
 'pandas>=1.0.3,<2.0.0',
 'poetry_version>=0.1.5,<0.2.0',
 'pytest-datadir-mgr>=1.0.1,<2.0.0',
 'seaborn>=0.10.1,<0.11.0',
 'sh>=1.13.1,<2.0.0']

entry_points = \
{'console_scripts': ['azulejo = azulejo:cli']}

setup_kwargs = {
    'name': 'azulejo',
    'version': '0.6.2',
    'description': 'tile phylogenetic space with subtrees',
    'long_description': '.. epigraph:: azulejo\n              noun INFORMAL\n              a glazed tile, usually blue, found on the inside of churches and palaces in Spain and Portugal.\n\n\nazulejo tiles phylogenetic space with subtrees\n\n\n+-------------------+------------+------------+\n| Latest Release    | |pypi|     | |azulejo|  |\n+-------------------+------------+            +\n| GitHub            | |repo|     |            |\n+-------------------+------------+            +\n| License           | |license|  |            |\n+-------------------+------------+            +\n| Travis Build      | |travis|   |            |\n+-------------------+------------+            +\n| Coverage          | |coverage| |            |\n+-------------------+------------+            +\n| Code Grade        | |codacy|   |            |\n+-------------------+------------+            +\n| Dependencies      | |depend|   |            |\n+-------------------+------------+            +\n| Issues            | |issues|   |            |\n+-------------------+------------+------------+\n\n\n.. |azulejo| image:: docs/azulejo.jpg\n     :target: https://en.wikipedia.org/wiki/Azulejo\n     :alt: azulejo Definition\n\n.. |pypi| image:: https://img.shields.io/pypi/v/azulejo.svg\n    :target: https://pypi.python.org/pypi/azulejo\n    :alt: Python package\n\n.. |repo| image:: https://img.shields.io/github/commits-since/legumeinfo/azulejo/0.11.svg\n    :target: https://github.com/legumeinfo/azulejo\n    :alt: GitHub repository\n\n.. |license| image:: https://img.shields.io/badge/License-BSD%203--Clause-blue.svg\n    :target: https://github.com/legumeinfo/azulejo/blob/master/LICENSE\n    :alt: License terms\n\n.. |rtd| image:: https://readthedocs.org/projects/azulejo/badge/?version=latest\n    :target: http://azulejo.readthedocs.io/en/latest/?badge=latest\n    :alt: Documentation Server\n\n.. |travis| image:: https://img.shields.io/travis/legumeinfo/azulejo.svg\n    :target:  https://travis-ci.org/legumeinfo/azulejo\n    :alt: Travis CI\n\n.. |codacy| image:: https://api.codacy.com/project/badge/Grade/99549f0ed4e6409e9f5e80a2c4bd806b\n    :target: https://www.codacy.com/app/joelb123/azulejo?utm_source=github.com&amp;utm_medium=referral&amp;utm_content=legumeinfo/azulejo&amp;utm_campaign=Badge_Grade\n    :alt: Codacy.io grade\n\n.. |coverage| image:: https://codecov.io/gh/legumeinfo/azulejo/branch/master/graph/badge.svg\n    :target: https://codecov.io/gh/legumeinfo/azulejo\n    :alt: Codecov.io test coverage\n\n.. |issues| image:: https://img.shields.io/github/issues/LegumeFederation/lorax.svg\n    :target:  https://github.com/legumeinfo/azulejo/issues\n    :alt: Issues reported\n\n.. |requires| image:: https://requires.io/github/legumeinfo/azulejo/requirements.svg?branch=master\n     :target: https://requires.io/github/legumeinfo/azulejo/requirements/?branch=master\n     :alt: Requirements Status\n\n.. |depend| image:: https://api.dependabot.com/badges/status?host=github&repo=legumeinfo/azulejo\n     :target: https://app.dependabot.com/accounts/legumeinfo/repos/203668510\n     :alt: dependabot dependencies\n',
    'author': 'Joel Berendzen',
    'author_email': 'joelb@ncgr.org',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/legumeinfo/azulejo',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6.1,<4.0.0',
}


setup(**setup_kwargs)



# -*- coding: utf-8 -*-
"""
Core logic for computing subtrees
"""

# standard library imports
import contextlib
import os
import sys
from collections import Counter
from collections import OrderedDict
from datetime import datetime
from itertools import chain
from itertools import combinations

# third-party imports
import click
import dask.bag as db
import networkx as nx
import numpy as np
import pandas as pd
from Bio import SeqIO
from Bio.Data import IUPACData
from dask.diagnostics import ProgressBar

# first-party imports
from loguru import logger

# module imports
from . import cli
from . import click_loguru
from .common import *

#
# global constants
#
UNITS = {
    "Mb": {"factor": 1, "outunits": "MB"},
    "Gb": {"factor": 1024, "outunits": "MB"},
    "s": {"factor": 1, "outunits": "s"},
    "m": {"factor": 60, "outunits": "s"},
    "h": {"factor": 3600, "outunits": "s"},
}
SEQ_IN_LINE = 6
IDENT_STATS_LINE = 7
FIRST_LOG_LINE = 14
LAST_LOG_LINE = 23
STAT_SUFFIXES = ["siz", "mem", "time", "memory"]
RENAME_STATS = {
    "throughput": "throughput_seq_s",
    "time": "CPU_time",
    "max_size": "max_cluster_size",
    "avg_size": "avg_cluster_size",
    "min_size": "min_cluster_size",
    "seqs": "unique_seqs",
    "singletons": "singleton_clusters",
}
ID_SEPARATOR = "."
IDENT_LOG_MIN = -3
IDENT_LOG_MAX = 0
EPSILON = 0.000001
FILETYPE = "pdf"
MAX_BINS = 10
DEFAULT_STEPS = 16
ALPHABET = IUPACData.protein_letters + "X" + "-"
#
# Classes
#
class ElapsedTimeReport:
    def __init__(self, name):
        self.start_time = datetime.now()
        self.name = name

    def elapsed(self, next_name):
        now = datetime.now()
        seconds = (now - self.start_time).total_seconds()
        report = f"{self.name} phase took {seconds:d} seconds"
        self.start_time = now
        self.name = next_name
        return report


class Sanitizer(object):
    """clean up and count potential problems with sequence

       potential problems are:
          dashes:    (optional, removed if remove_dashes=True)
          alphabet:  if not in IUPAC set, changed to 'X'
    """

    def __init__(self, remove_dashes=False):
        self.remove_dashes = remove_dashes
        self.seqs_sanitized = 0
        self.chars_in = 0
        self.chars_removed = 0
        self.chars_fixed = 0
        self.endchars_removed = 0

    def char_remover(self, s, character):
        """remove positions with a given character

        :param s: mutable sequence
        :return: sequence with characters removed
        """
        removals = [i for i, j in enumerate(s) if j == character]
        self.chars_removed += len(removals)
        [s.pop(pos - k) for k, pos in enumerate(removals)]
        return s

    def fix_alphabet(self, s):
        """replace everything out of alphabet with 'X'

        :param s: mutable sequence, upper-cased
        :return: fixed sequence
        """
        fix_positions = [pos for pos, char in enumerate(s) if char not in ALPHABET]
        self.chars_fixed = len(fix_positions)
        [s.__setitem__(pos, "X") for pos in fix_positions]
        return s

    def remove_char_on_ends(self, s, character):
        """remove leading/trailing ambiguous residues

        :param s: mutable sequence
        :return: sequence with characterss removed from ends
        """
        in_len = len(s)
        while s[-1] == character:
            s.pop()
        while s[0] == character:
            s.pop(0)
        self.endchars_removed += in_len - len(s)
        return s

    def sanitize(self, s):
        """sanitize alphabet use while checking lengths

        :param s: mutable sequence
        :return: sanitized sequence
        """
        self.seqs_sanitized += 1
        self.chars_in += len(s)
        if len(s) and self.remove_dashes:
            s = self.char_remover(s, "-")
        if len(s):
            s = self.fix_alphabet(s)
        if len(s):
            s = self.remove_char_on_ends(s, "X")
        return s


def read_synonyms(filepath):
    synonym_dict = {}
    try:
        df = pd.read_csv(filepath, sep="\t")
    except FileNotFoundError:
        logger.error(f'Synonym tsv file "{substrpath}" does not exist')
        sys.exit(1)
    except pd.errors.EmptyDataError:
        logger.error(f'Synonym tsv "{substrpath}" is empty')
        sys.exit(1)
    if len(df):
        if "#file" in df:
            df.drop("#file", axis=1, inplace=True)
        key = list(set(("Substr", "Dups")).intersection(set(df.columns)))[0]
        for group in df.groupby("id"):
            synonym_dict[group[0]] = [substr for substr in group[1][key]]
    return synonym_dict


def parse_usearch_log(filepath, rundict):
    with filepath.open() as logfile:
        for lineno, line in enumerate(logfile):
            if lineno < FIRST_LOG_LINE:
                if lineno == SEQ_IN_LINE:
                    split = line.split()
                    rundict["seqs_in"] = int(split[0])
                    rundict["singleton_seqs_in"] = int(split[4])
                if lineno == IDENT_STATS_LINE:
                    split = line.split()
                    rundict["max_identical_seqs"] = int(split[6].rstrip(","))
                    rundict["avg_identical_seqs"] = float(split[8])
                continue
            if lineno > LAST_LOG_LINE:
                break
            split = line.split()
            if split:
                stat = split[0].lower()
                if split[1] in STAT_SUFFIXES:
                    stat += "_" + split[1]
                    val = split[2]
                else:
                    val = split[1].rstrip(",")
                # rename poorly-named stats
                if stat in RENAME_STATS:
                    stat = RENAME_STATS[stat]
                # strip stats with units at the end
                conversion_factor = 1
                for unit in UNITS.keys():
                    if val.endswith(unit):
                        val = val.rstrip(unit)
                        conversion_factor = UNITS[unit]["factor"]
                        stat += "_" + UNITS[unit]["outunits"]
                        break
                # convert string values to int or float where possible
                try:
                    val = int(val)
                    val *= conversion_factor
                except:
                    try:
                        val = float(val)
                        val *= conversion_factor
                    except:
                        pass
                rundict[stat] = val


@contextlib.contextmanager
def in_working_directory(path):
    """Changes working directory and returns to previous on exit."""
    original_cwd = Path.cwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(original_cwd)


def get_fasta_ids(fasta):
    idset = set()
    with fasta.open() as f:
        for line in f:
            if line.startswith(">"):
                idset.add(line.split()[0][1:])
    return list(idset)


def parse_chromosome(id):
    #
    # If id contains an underscore, work on the
    # last part only (e.g., MtrunA17_Chr4g0009691)
    #
    undersplit = id.split("_")
    if len(undersplit) > 1:
        id = undersplit[-1].upper()
        if id.startswith("CHR"):
            id = id[3:]
    #
    # Chromosome numbers are integers suffixed by 'G'
    #
    try:
        chr = "Chr" + str(int(id[: id.index("G")]))
    except:
        chr = None
    return chr


def parse_subids(id):
    subids = id.split(ID_SEPARATOR)
    subids += [chr for chr in [parse_chromosome(id) for id in subids] if chr is not None]
    return subids


def parse_clusters(outdir, identity, delete=True, count_clusters=True, synonyms={}):
    cluster_list = []
    id_list = []
    degree_list = []
    size_list = []
    degree_counter = Counter()
    any_counter = Counter()
    all_counter = Counter()
    graph = nx.Graph()
    for fasta in outdir.glob("*"):
        cluster_id = int(fasta.name)
        ids = get_fasta_ids(fasta)
        if len(synonyms):
            syn_ids = set(ids).intersection(synonyms.keys())
            [ids.extend(synonyms[i]) for i in syn_ids]
        n_ids = len(ids)
        degree_list.append(n_ids)
        degree_counter.update({n_ids: 1})
        id_list += ids
        cluster_list += [cluster_id] * n_ids
        size_list += [n_ids] * n_ids
        #
        # Do 'any' and 'all' counters
        #
        id_counter = Counter()
        id_counter.update(chain.from_iterable([parse_subids(id) for id in ids]))
        if count_clusters:
            any_counter.update(id_counter.keys())
            all_counter.update([id for id in id_counter.keys() if id_counter[id] == n_ids])
        elif n_ids > 1:
            any_counter.update({s: n_ids for s in id_counter.keys()})
            all_counter.update({id: n_ids for id in id_counter.keys() if id_counter[id] == n_ids})
        #
        # Do graph components
        #
        graph.add_nodes_from(ids)
        if n_ids > 1:
            edges = combinations(ids, 2)
            graph.add_edges_from(edges, weight=n_ids)
        if delete:
            fasta.unlink()
    if delete:
        outdir.rmdir()
    return graph, cluster_list, id_list, size_list, degree_list, degree_counter, any_counter, all_counter


def prettyprint_float(x, digits):
    format_string = "%." + f"{digits:d}" + "f"
    return (format_string % x).rstrip("0").rstrip(".")


def cluster_set_name(stem, identity):
    """Get a setname that identity."""
    if identity == 1.0:
        digits = "10000"
    else:
        digits = (f"{identity:.4f}")[2:]
    return f"{stem}-nr-{digits}"


@cli.command()
@click_loguru.init_logger()
@click.argument("seqfile")
@click.option("--identity", "-i", default=0.0, help="Minimum sequence identity (float, 0-1). [default: lowest]")
@click.option("--min_id_freq", "-m", default=0, show_default=True, help="Minimum frequency of ID components.")
@click.option(
    "--delete/--no-delete",
    "-d/-n",
    is_flag=True,
    default=True,
    help="Delete primary output of clustering. [default: delete]",
)
@click.option(
    "--write_ids/--no-write_ids",
    "-w",
    is_flag=True,
    default=False,
    help="Write file of ID-to-clusters. [default: delete]",
)
@click.option(
    "--do_calc/--no-do_calc",
    "-c/-x",
    is_flag=True,
    default=True,
    help="Write file of ID-to-clusters. [default: do_calc]",
)
@click.option("--substrs", help="subpath to file of substrings. [default: none]")
@click.option("--dups", help="subpath to file of duplicates. [default: none]")
def usearch_cluster(
    seqfile, identity, delete=True, write_ids=False, do_calc=True, min_id_freq=0, substrs=None, dups=None
):
    """Cluster above a global sequence identity threshold"""
    from sh import usearch

    try:
        inpath, dirpath = get_paths_from_file(seqfile)
    except FileNotFoundError:
        logger.error(f'Input file "{seqfile}" does not exist!')
        sys.exit(1)
    stem = inpath.stem
    dirpath = inpath.parent
    outname = cluster_set_name(stem, identity)
    outdir = f"{outname}/"
    logfile = f"{outname}.log"
    outfilepath = dirpath / outdir
    logfilepath = dirpath / logfile
    histfilepath = dirpath / (f"{outname}-degreedist.tsv")
    gmlfilepath = dirpath / (f"{outname}.gml")
    statfilepath = dirpath / (f"{outname}-stats.tsv")
    anyfilepath = dirpath / (f"{outname}-anyhist.tsv")
    allfilepath = dirpath / (f"{outname}-allhist.tsv")
    idpath = dirpath / (f"{outname}-ids.tsv")
    logger.info(f"%{prettyprint_float(identity * 100, 2)}% sequence identity output to {outname}*")
    if not delete:
        logger.debug(f"Cluster files will be kept in {logfile} and {outdir}")
    if write_ids:
        logger.debug(f"File of cluster ID usage will be written to {anyfilepath} and {allfilepath}")
    if not do_calc:
        if not logfilepath.exists():
            logger.error("Previous results must exist, rerun with --do_calc")
            sys.exit(1)
        logger.debug("Using previous results for calculation")
    if min_id_freq:
        logger.debug(f"Minimum number of times ID's must occur to be counted: {min_id_freq}")
    synonyms = {}
    if substrs is not None:
        logger.debug(f"using duplicates in {dirpath / dups}")
        synonyms.update(read_synonyms(dirpath / substrs))
    if dups is not None:
        logger.debug(f"using duplicates in {dirpath/dups}")
        synonyms.update(read_synonyms(dirpath / dups))
    timer = ElapsedTimeReport("usearch")
    if do_calc:
        #
        # Delete previous results, if any.
        #
        if outfilepath.exists() and outfilepath.is_file():
            outfilepath.unlink()
        elif outfilepath.exists() and outfilepath.is_dir():
            for file in outfilepath.glob("*"):
                file.unlink()
                pass
        else:
            outfilepath.mkdir()
        #
        # Do the calculation.
        #
        with in_working_directory(dirpath):
            output = usearch(["-cluster_fast", seqfile, "-id", identity, "-clusters", outdir, "-log", logfile])
            logger.debug(output)
            logger.debug(timer.elapsed("parse"))
    run_stat_dict = OrderedDict([("divergence", 1.0 - identity)])
    parse_usearch_log(logfilepath, run_stat_dict)
    run_stats = pd.DataFrame(list(run_stat_dict.items()), columns=["stat", "val"])
    run_stats.set_index("stat", inplace=True)
    run_stats.to_csv(statfilepath, sep="\t")
    if delete:
        logfilepath.unlink()
    cluster_graph, clusters, ids, sizes, degrees, degree_counts, any_counts, all_counts = parse_clusters(
        outfilepath, identity, delete=delete, synonyms=synonyms
    )
    #
    # Write out list of clusters and ids.
    #
    id_frame = pd.DataFrame.from_dict({"id": ids, "cluster": clusters, "siz": sizes})
    id_frame.sort_values("siz", ascending=False, inplace=True)
    id_frame = id_frame.reindex(["cluster", "siz", "id",], axis=1)
    id_frame.reset_index(inplace=True)
    id_frame.drop(["index"], axis=1, inplace=True)
    id_frame.to_csv(idpath, sep="\t")
    del ids, clusters, sizes, id_frame
    logger.debug(timer.elapsed("graph"))
    #
    # Write out degree distribution.
    #
    cluster_hist = pd.DataFrame(list(degree_counts.items()), columns=["degree", "clusters"])
    # cluster_hist['degree'] = cluster_hist['degree'] - 1
    cluster_hist.sort_values(["degree"], inplace=True)
    cluster_hist.set_index("degree", inplace=True)
    total_clusters = cluster_hist["clusters"].sum()
    cluster_hist["pct_total"] = cluster_hist["clusters"] * 100.0 / total_clusters
    cluster_hist.to_csv(histfilepath, sep="\t", float_format="%06.3f")
    del degree_counts
    #
    # Do histograms of "any" and "all" id usage in cluster
    #
    hist_value = f"{identity:f}"
    any_hist = pd.DataFrame(list(any_counts.items()), columns=["id", hist_value])
    any_hist.set_index("id", inplace=True)
    any_hist.sort_values(hist_value, inplace=True, ascending=False)
    all_hist = pd.DataFrame(list(all_counts.items()), columns=["id", hist_value])
    all_hist.set_index("id", inplace=True)
    all_hist.sort_values(hist_value, inplace=True, ascending=False)
    if min_id_freq:
        any_hist = any_hist[any_hist[hist_value] > min_id_freq]
        all_hist = all_hist[all_hist[hist_value] > min_id_freq]
    if write_ids:
        any_hist.to_csv(anyfilepath, sep="\t")
        all_hist.to_csv(allfilepath, sep="\t")
    #
    # Compute cluster stats
    #
    # degree_sequence = sorted([d for n, d in cluster_graph.degree()], reverse=True)
    # degreeCount = Counter(degree_sequence)
    # degree_hist = pd.DataFrame(list(degreeCount.items()),
    #                           columns=['degree', 'count'])
    # degree_hist.set_index('degree', inplace=True)
    # degree_hist.sort_values('degree', inplace=True)
    # degree_hist.to_csv(histfilepath, sep='\t')
    nx.write_gml(cluster_graph, gmlfilepath)
    logger.debug(timer.elapsed("final"))
    return run_stat_dict, cluster_graph, cluster_hist, any_hist, all_hist


@cli.command()
@click_loguru.init_logger()
@click.argument("seqfile")
@click.option("--steps", "-s", default=DEFAULT_STEPS, show_default=True, help="# of steps from lowest to highest")
@click.option("--min_id_freq", "-m", default=0, show_default=True, help="Minimum frequency of ID components.")
@click.option("--substrs", help="subpath to file of substrings. [default: none]")
@click.option("--dups", help="subpath to file of duplicates. [default: none]")
def cluster_in_steps(seqfile, steps, min_id_freq=0, substrs=None, dups=None):
    """Cluster in steps from low to 100% sequence identity"""
    try:
        inpath, dirpath = get_paths_from_file(seqfile)
    except FileNotFoundError:
        logger.error('Input file "%s" does not exist!', seqfile)
        sys.exit(1)
    stat_path = dirpath / (inpath.stem + STATFILE_SUFFIX)
    any_path = dirpath / (inpath.stem + ANYFILE_SUFFIX)
    all_path = dirpath / (inpath.stem + ALLFILE_SUFFIX)
    logsteps = [1.0] + list(1.0 - np.logspace(IDENT_LOG_MIN, IDENT_LOG_MAX, num=steps))
    logger.info(
        "Clustering at %d levels from %s%% to %s%% global sequence identity",
        steps,
        prettyprint_float(min(logsteps) * 100.0, 2),
        prettyprint_float(max(logsteps) * 100.0, 2),
    )
    stat_list = []
    all_frames = []
    any_frames = []
    for id_level in logsteps:
        stats, graph, hist, any, all = usearch_cluster.callback(
            seqfile, id_level, min_id_freq=min_id_freq, substrs=substrs, dups=dups
        )
        stat_list.append(stats)
        any_frames.append(any)
        all_frames.append(all)
    logger.info("Collating results on %s", seqfile)
    #
    # Concatenate and write stats
    #
    stats = pd.DataFrame(stat_list)
    stats.to_csv(stat_path, sep="\t")
    #
    # Concatenate any/all data
    #
    any = pd.concat(any_frames, axis=1, join="inner", sort=True, ignore_index=False)
    any.to_csv(any_path, sep="\t")
    all = pd.concat(all_frames, axis=1, join="inner", sort=True, ignore_index=False)
    all.to_csv(all_path, sep="\t")


@cli.command()
@click_loguru.init_logger(logfile=False)
@click.argument("infile")
def clusters_to_histograms(infile):
    """Compute histograms from a tab-delimited cluster file"""
    try:
        inpath, dirpath = get_paths_from_file(infile)
    except FileNotFoundError:
        logger.error('Input file "%s" does not exist!', infile)
        sys.exit(1)
    histfilepath = dirpath / (inpath.stem + "-sizedist.tsv")
    clusters = pd.read_csv(dirpath / infile, sep="\t", index_col=0)
    cluster_counter = Counter()
    for cluster_id, group in clusters.groupby(["cluster"]):
        cluster_counter.update({len(group): 1})
    logger.info("writing to %s", histfilepath)
    cluster_hist = pd.DataFrame(list(cluster_counter.items()), columns=["siz", "clusts"])
    total_clusters = cluster_hist["clusts"].sum()
    cluster_hist["%clusts"] = cluster_hist["clusts"] * 100.0 / total_clusters
    cluster_hist["%genes"] = cluster_hist["clusts"] * cluster_hist["siz"] * 100.0 / len(clusters)
    cluster_hist.sort_values(["siz"], inplace=True)
    cluster_hist.set_index("siz", inplace=True)
    cluster_hist.to_csv(histfilepath, sep="\t", float_format="%06.3f")


@cli.command()
@click_loguru.init_logger(logfile=False)
@click.argument("file1")
@click.argument("file2")
def compare_clusters(file1, file2):
    """ compare one cluster file with another
    """
    path1 = Path(file1)
    path2 = Path(file2)
    commondir = Path(os.path.commonpath([path1, path2]))
    missing1 = commondir / "notin1.tsv"
    missing2 = commondir / "notin2.tsv"
    clusters1 = pd.read_csv(path1, sep="\t", index_col=0)
    print(f"{len(clusters1):d} members in {file1}")
    clusters2 = pd.read_csv(path2, sep="\t", index_col=0)
    print(f"{len(clusters2):d} members in {file2}")
    ids1 = set(clusters1["id"])
    ids2 = set(clusters2["id"])
    notin1 = pd.DataFrame(ids2.difference(ids1), columns=["id"])
    notin1.sort_values("id", inplace=True)
    notin1.to_csv(missing1, sep="\t")
    notin2 = pd.DataFrame(ids1.difference(ids2), columns=["id"])
    notin2.sort_values("id", inplace=True)
    notin2.to_csv(missing2, sep="\t")

    print("%d ids not in ids1" % len(notin1))
    print("%d ids not in ids2" % len(notin2))
    print(f"{len(clusters1):d} in {file1} after dropping")


@cli.command()
@click_loguru.init_logger(logfile=False)
@click.argument("setname")
@click.argument("filelist", nargs=-1)
def scanfiles(setname, filelist):
    """scan a set of files, producing summary statistics"""
    setpath = Path(setname)
    if len(filelist) < 1:
        logger.error("Empty FILELIST, aborting")
        sys.exit(1)
    if setpath.exists() and setpath.is_file():
        setpath.unlink()
    elif setpath.exists() and setpath.is_dir():
        logger.debug("set path %s exists", setname)
    else:
        logger.info('creating output directory "%s/"', setname)
        setpath.mkdir()
    outfile = setpath / "all.faa"
    statfile = setpath / "records.tsv"
    out_sequences = []
    lengths = []
    ids = []
    files = []
    positions = []
    for file in filelist:
        position = 0
        logger.info("scanning %s", file)
        sanitizer = Sanitizer(remove_dashes=True)
        with open(file, "rU") as handle:
            for record in SeqIO.parse(handle, SEQ_FILE_TYPE):
                seq = record.seq.upper().tomutable()
                seq = sanitizer.sanitize(seq)
                if not len(seq):
                    # zero-length string after sanitizing
                    continue
                record.seq = seq.toseq()
                ids.append(record.id)
                lengths.append(len(record))
                out_sequences.append(record)
                files.append(file)
                positions.append(position)
                position += 1
    logger.debug("writing output files")
    with outfile.open("w") as output_handle:
        SeqIO.write(out_sequences, output_handle, SEQ_FILE_TYPE)
    df = pd.DataFrame(list(zip(files, ids, positions, lengths)), columns=["file", "id", "pos", "len"])
    df.to_csv(statfile, sep="\t")


@cli.command()
@click_loguru.init_logger(logfile=False)
@click.argument("infile")
@click.argument("recordfile")
def add_singletons(infile, recordfile):
    """Add singleton clusters to cluster file"""
    clusters = pd.read_csv(infile, header=None, names=["cluster", "id"], sep="\t")
    records = pd.read_csv(recordfile, sep="\t")
    id_set = set(records["id"])
    records.drop(["file", "pos"], axis=1, inplace=True)
    records.set_index("id", inplace=True)
    records.drop(["Unnamed: 0"], axis=1, inplace=True)
    len_dict = records.to_dict("id")
    del records
    outfile = "clusters.tsv"
    ids = []
    cluster_ids = []
    sizes = []
    lengths = []
    n_clusters = 0
    grouping = clusters.groupby("cluster").size().sort_values(ascending=False)
    for idx in grouping.index:
        size = grouping.loc[idx]
        cluster = clusters.loc[(clusters["cluster"] == idx)]
        for gene_id in cluster["id"]:
            id_set.remove(gene_id)
            ids.append(gene_id)
            cluster_ids.append(n_clusters)
            sizes.append(size)
            lengths.append(len_dict[gene_id]["len"])
        n_clusters += 1
    logger.info("%s singletons", len(id_set))
    for singleton in id_set:
        cluster_ids.append(n_clusters)
        ids.append(singleton)
        sizes.append(1)
        lengths.append(len_dict[singleton]["len"])
        n_clusters += 1
    df = pd.DataFrame(list(zip(cluster_ids, ids, sizes, lengths)), columns=["cluster", "id", "siz", "len"])
    df.to_csv(outfile, sep="\t")


@cli.command()
@click_loguru.init_logger(logfile=False)
@click.argument("infile")
def adjacency_to_graph(infile):
    gmlfilepath = "synteny.gml"
    clusters = pd.read_csv(infile, sep="\t", index_col=0)
    graph = nx.Graph()
    grouping = clusters.groupby("cluster").size().sort_values(ascending=False)
    for idx in grouping.index:
        size = grouping.loc[idx]
        cluster = clusters.loc[(clusters["cluster"] == idx)]
        ids = list(cluster["id"])
        #
        # Do graph components
        #
        graph.add_nodes_from(ids)
        if len(ids) > 1:
            edges = combinations(ids, 2)
            graph.add_edges_from(edges, weight=len(ids))
    nx.write_gml(graph, gmlfilepath)


def compute_subclusters(cluster, cluster_size_dict=None):
    #
    # compute a dictionary of per-subcluster stats first,
    # so subclusters can be ordered by length
    #
    subcl_frame = pd.DataFrame(
        [
            {"homo_id": i, "mean_len": g["len"].mean(), "std": g["len"].std(), "sub_siz": len(g)}
            for i, g in cluster.groupby("link")
        ]
    )
    subcl_frame["link"] = subcl_frame["homo_id"]
    subcl_frame["cont"] = subcl_frame["sub_siz"] == [cluster_size_dict[id] for id in subcl_frame["homo_id"]]
    subcl_frame.loc[subcl_frame["cont"], "link"] = np.nan
    subcl_frame.sort_values(["sub_siz", "std", "mean_len"], ascending=[False, True, False], inplace=True)
    subcl_frame["sub"] = list(range(len(subcl_frame)))
    subcl_frame.index = list(range(len(subcl_frame)))
    # normalized length is NaN for first element
    subcl_frame["norm"] = [np.nan] + list(subcl_frame["mean_len"][1:] / subcl_frame["mean_len"][0])
    subcl_dict = subcl_frame.set_index("homo_id").to_dict("index")
    #
    # subcluster attributes are done, now copy them into the cluster frame
    #
    for attr in ["norm", "std", "sub_siz", "sub", "link"]:
        cluster[attr] = [subcl_dict[id][attr] for id in cluster["link"]]
    if subcl_frame["cont"].all():
        cluster["cont"] = 1
    else:
        cluster["cont"] = 0
    return cluster


@cli.command()
@click_loguru.init_logger(logfile=False)
@click.option("--first_n", default=0, show_default=True, help="Only process this many clusters.")
@click.option("--clust_size", default=0, show_default=True, help="Process only clusters of this size.")
@click.option("--parallel/--no-parallel", is_flag=True, default=True, show_default=True, help="Process in parallel.")
@click.option("-q", "--quiet", is_flag=True, show_default=True, default=False, help="Suppress logging to stderr.")
@click.argument("synfile")
@click.argument("homofile")
def combine_clusters(first_n, clust_size, synfile, homofile, quiet, parallel):
    """combine synteny and homology clusters"""
    timer = ElapsedTimeReport("reading/preparing")
    syn = pd.read_csv(synfile, sep="\t", index_col=0)
    homo = pd.read_csv(homofile, sep="\t", index_col=0)
    homo_id_dict = dict(zip(homo.id, homo.cluster))
    cluster_size_dict = dict(zip(homo.cluster, homo.siz))
    cluster_count = 0
    arg_list = []
    if first_n:
        logger.debug(f"processing only first {first_n:d} clusters")
    if not quiet and clust_size:
        logger.info("only clusters of size %d will be used", clust_size)
    syn["link"] = [homo_id_dict[id] for id in syn["id"]]
    for cluster_id, gr in syn.groupby(["cluster"]):
        cl = gr.copy()  # copy, so mutable
        clsize = cl["siz"].iloc[0]
        if clust_size and (clsize != clust_size):
            continue
        if first_n and cluster_count == first_n:
            break
        arg_list.append(cl)
        cluster_count += 1
    del syn
    if parallel:
        bag = db.from_sequence(arg_list)
    else:
        cluster_list = []
    if not quiet:
        logger.info("Combining %d synteny/homology clusters:", cluster_count)
        ProgressBar().register()
    if parallel:
        cluster_list = bag.map(compute_subclusters, cluster_size_dict=cluster_size_dict)
    else:
        for clust in arg_list:
            cluster_list.append(compute_subclusters(clust, cluster_size_dict))
    logger.debug(timer.elapsed("concatenating/writing clusters"))
    out_frame = pd.concat(cluster_list)
    out_frame.sort_values(["cluster", "sub"], inplace=True)
    out_frame.index = list(range(len(out_frame)))
    for column in ["sub", "sub_siz"]:
        out_frame[column] = out_frame[column].astype(int)
    out_frame["link"] = out_frame["link"].map(lambda x: "" if pd.isnull(x) else f"{x:.0f}")
    out_frame.to_csv(
        "combined.tsv",
        sep="\t",
        columns=["cluster", "siz", "sub", "sub_siz", "cont", "norm", "std", "len", "link", "id"],
        float_format="%.3f",
        index=False,
    )
    logger.debug(timer.elapsed("computing stats"))
    n_fully_contained = len(set(out_frame[out_frame["cont"] == 1]["cluster"]))
    logger.info(
        "%d of %d clusters are fully contained (%.1f%%)",
        n_fully_contained,
        cluster_count,
        n_fully_contained * 100 / cluster_count,
    )
    # print('subclusters\tN\t%clusts')
    # for count in sorted(contained_subclst_count.keys()):
    #    print('%d\t%d\t%.1f'%(count, contained_subclst_count[count],
    #                         contained_subclst_count[count]*100/cluster_count))
    # uncontained = cluster_count - n_fully_contained
    # print('%d of %d clusters are unontained (%.1f%%)'
    #      % (uncontained, cluster_count,
    #         uncontained * 100 / cluster_count))
